# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Tuple

import mxnet as mx

from gluonts.core.component import validated
from gluonts.mx import Tensor
from gluonts.mx.block.scaler import MeanScaler, NOPScaler
from gluonts.mx.distribution import DistributionOutput
from gluonts.mx.util import weighted_average


class SimpleFeedForwardNetworkBase(mx.gluon.HybridBlock):
    """
    Abstract base class to implement feed-forward networks for probabilistic
    time series prediction.

    This class does not implement hybrid_forward: this is delegated
    to the two subclasses SimpleFeedForwardTrainingNetwork and
    SimpleFeedForwardPredictionNetwork, that define respectively how to
    compute the loss and how to generate predictions.

    Parameters
    ----------
    num_hidden_dimensions
        Number of hidden nodes in each layer.
    prediction_length
        Number of time units to predict.
    context_length
        Number of time units that condition the predictions.
    batch_normalization
        Whether to use batch normalization.
    mean_scaling
        Scale the network input by the data mean and the network output by
        its inverse.
    distr_output
        Distribution to fit.
    kwargs
    """

    # Needs the validated decorator so that arguments types are checked and
    # the block can be serialized.
    @validated()
    def __init__(
        self,
        num_hidden_dimensions: List[int],
        prediction_length: int,
        context_length: int,
        batch_normalization: bool,
        mean_scaling: bool,
        distr_output: DistributionOutput,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.num_hidden_dimensions = num_hidden_dimensions
        self.prediction_length = prediction_length
        self.context_length = context_length
        self.batch_normalization = batch_normalization
        self.mean_scaling = mean_scaling
        self.distr_output = distr_output

        with self.name_scope():
            self.distr_args_proj = self.distr_output.get_args_proj()
            self.mlp = mx.gluon.nn.HybridSequential()
            dims = self.num_hidden_dimensions
            for layer_no, units in enumerate(dims[:-1]):
                self.mlp.add(mx.gluon.nn.Dense(units=units, activation="relu"))
                if self.batch_normalization:
                    self.mlp.add(mx.gluon.nn.BatchNorm())
            self.mlp.add(mx.gluon.nn.Dense(units=prediction_length * dims[-1]))
            self.mlp.add(
                mx.gluon.nn.HybridLambda(
                    lambda F, o: F.reshape(
                        o, (-1, prediction_length, dims[-1])
                    )
                )
            )
            self.scaler = MeanScaler() if mean_scaling else NOPScaler()

    def get_distr_args(
        self, F, past_target: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor]:
        """
        Given past target values, applies the feed-forward network and maps the
        output to the parameter of probability distribution for future
        observations.

        Parameters
        ----------
        F
        past_target
            Tensor containing past target observations.
            Shape: (batch_size, context_length, target_dim).

        Returns
        -------
        Tensor
            The parameters of distribution.
        Tensor
            An array containing the location (shift) of the distribution.
        Tensor
            An array containing the scale of the distribution.
        """
        scaled_target, target_scale = self.scaler(
            past_target,
            F.ones_like(past_target),
        )
        mlp_outputs = self.mlp(scaled_target)
        distr_args = self.distr_args_proj(mlp_outputs)
        scale = target_scale.expand_dims(axis=1)
        loc = F.zeros_like(scale)
        return distr_args, loc, scale


class SimpleFeedForwardTrainingNetwork(SimpleFeedForwardNetworkBase):
    def hybrid_forward(
        self,
        F,
        past_target: Tensor,
        future_target: Tensor,
        future_observed_values: Tensor,
    ) -> Tensor:
        """
        Computes a probability distribution for future data given the past, and
        returns the loss associated with the actual future observations.

        Parameters
        ----------
        F
        past_target
            Tensor with past observations.
            Shape: (batch_size, context_length, target_dim).
        future_target
            Tensor with future observations.
            Shape: (batch_size, prediction_length, target_dim).
        future_observed_values
            Tensor indicating which values in the target are observed, and
            which ones are imputed instead.

        Returns
        -------
        Tensor
            Loss tensor. Shape: (batch_size, ).
        """
        distr_args, loc, scale = self.get_distr_args(F, past_target)
        distr = self.distr_output.distribution(
            distr_args, loc=loc, scale=scale
        )

        # (batch_size, prediction_length, target_dim)
        loss = distr.loss(future_target)

        weighted_loss = weighted_average(
            F=F, x=loss, weights=future_observed_values, axis=1
        )

        # (batch_size, )
        return weighted_loss


class SimpleFeedForwardSamplingNetwork(SimpleFeedForwardNetworkBase):
    @validated()
    def __init__(
        self, num_parallel_samples: int = 100, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.num_parallel_samples = num_parallel_samples

    def hybrid_forward(self, F, past_target: Tensor) -> Tensor:
        """
        Computes a probability distribution for future data given the past, and
        draws samples from it.

        Parameters
        ----------
        F
        past_target
            Tensor with past observations.
            Shape: (batch_size, context_length, target_dim).

        Returns
        -------
        Tensor
            Prediction sample. Shape: (batch_size, samples, prediction_length).
        """

        distr_args, loc, scale = self.get_distr_args(F, past_target)
        distr = self.distr_output.distribution(
            distr_args, loc=loc, scale=scale
        )

        # (num_samples, batch_size, prediction_length)
        samples = distr.sample(self.num_parallel_samples)

        # (batch_size, num_samples, prediction_length)
        return samples.swapaxes(0, 1)


class SimpleFeedForwardDistributionNetwork(SimpleFeedForwardNetworkBase):
    @validated()
    def __init__(
        self, num_parallel_samples: int = 100, *args, **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.num_parallel_samples = num_parallel_samples

    def hybrid_forward(self, F, past_target: Tensor) -> Tensor:
        """
        Computes the parameters of distribution for future data given the past,
        and draws samples from it.

        Parameters
        ----------
        F
        past_target
            Tensor with past observations.
            Shape: (batch_size, context_length, target_dim).

        Returns
        -------
        Tensor
            The parameters of distribution.
        Tensor
            An array containing the location (shift) of the distribution.
        Tensor
            An array containing the scale of the distribution.
        """
        distr_args, loc, scale = self.get_distr_args(F, past_target)
        return distr_args, loc, scale



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from .custom_binning import CustomBinning
from .dim_expansion import DimExpansion
from .discrete_pit import DiscretePIT
from .embedding import Embedding
from .global_relative_binning import GlobalRelativeBinning
from .hybrid_representation import HybridRepresentation
from .local_absolute_binning import LocalAbsoluteBinning
from .mean_scaling import MeanScaling
from .representation import Representation
from .representation_chain import RepresentationChain

__all__ = [
    "Representation",
    "CustomBinning",
    "GlobalRelativeBinning",
    "LocalAbsoluteBinning",
    "HybridRepresentation",
    "MeanScaling",
    "DimExpansion",
    "Embedding",
    "DiscretePIT",
    "RepresentationChain",
]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import ast
from functools import singledispatch


@singledispatch
def parse_expr(v):
    raise ValueError(v)


@parse_expr.register(ast.Num)
def parse_num(v: ast.Num):
    return v.n


@parse_expr.register(ast.UnaryOp)
def parse_unary_op(v: ast.UnaryOp):
    operand = parse_expr(v.operand)

    assert isinstance(operand, (float, int))
    assert isinstance(v.op, ast.USub)

    return -operand


@parse_expr.register(ast.Str)
def parse_str(v: ast.Str):
    return v.s


@parse_expr.register(ast.List)
def parse_list(v: ast.List):
    return list(map(parse_expr, v.elts))


@parse_expr.register(ast.Tuple)
def parse_tuple(v: ast.Tuple):
    return tuple(map(parse_expr, v.elts))


@parse_expr.register(ast.Dict)
def parse_dict(v: ast.Dict):
    keys = map(parse_expr, v.keys)
    values = map(parse_expr, v.values)
    return dict(zip(keys, values))


@parse_expr.register(ast.Set)
def parse_set(v: ast.Set):
    return set(map(parse_expr, v.elts))


@parse_expr.register(ast.keyword)
def parse_keyword(v: ast.keyword):
    return v.arg, parse_expr(v.value)


@parse_expr.register(ast.NameConstant)
def parse_name_constant(v: ast.NameConstant):
    return v.value


@parse_expr.register(ast.Constant)
def parse_constant(v: ast.Constant):
    return v.value


@parse_expr.register(ast.Attribute)
def parse_attribute(v: ast.Attribute, path=()):
    if isinstance(v.value, ast.Name):
        return {
            "__kind__": "type",
            "class": ".".join((v.value.id, v.attr) + path),
        }
    else:
        assert isinstance(v.value, ast.Attribute)
        return parse_attribute(v.value, (v.attr,) + path)


@parse_expr.register(ast.Name)
def parse_name(v: ast.Name):
    return {"__kind__": "type", "class": v.id}


@parse_expr.register(ast.Call)
def parse_expr_call(v: ast.Call):
    args = list(map(parse_expr, v.args))
    kwargs = dict(map(parse_keyword, v.keywords))

    class_name = parse_expr(v.func)["class"]
    obj = {"__kind__": "instance", "class": class_name}

    if args:
        obj["args"] = args

    if kwargs:
        obj["kwargs"] = kwargs

    return obj


def parse(s):
    module = ast.parse(s)
    expr = module.body[0]
    return parse_expr(expr.value)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
`gluonts.shell.sagemaker.dyn`
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Dynamic code loading for `gluonts.shell`.
"""


import os
import shutil
import subprocess
import sys
import tarfile
import tempfile
import zipfile
from functools import partial
from pathlib import Path
from typing import Optional

from gluonts.util import safe_extractall


class Installer:
    def __init__(self, packages):
        self.packages = packages
        self.cleanups = []

    def cleanup(self):
        for cleanup in self.cleanups:
            cleanup()

    def copy_install(self, path: Path):
        if path.is_file():
            shutil.copy(path, self.packages)
        elif path.is_dir():
            shutil.copytree(path, self.packages / path.name)

    def pip_install(self, path: Path):
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "--target",
                str(self.packages),
                str(path),
            ]
        )

    def install_requirement(self, path: Path):
        subprocess.check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "--target",
                str(self.packages),
                "--requirement",
                str(path),
            ]
        )

    def install(self, path):
        if path.is_file():
            if tarfile.is_tarfile(path):
                self.handle_archive(tarfile.open, safe_extractall, path)

            elif zipfile.is_zipfile(path):
                self.handle_archive(
                    zipfile.ZipFile, zipfile.ZipFile.extractall, path
                )

            elif path.match("requirements*.txt"):
                self.install_requirement(path)

            elif path.suffix == ".py":
                self.copy_install(path)

        elif path.is_dir():
            if (path / "setup.py").exists():
                self.pip_install(path)
            elif (path / "__init__.py").exists():
                self.copy_install(path)
            else:
                for subpath in path.iterdir():
                    self.install(subpath)

    def handle_archive(self, open_fn, extractall_fn, path):
        with open_fn(path) as archive:
            tempdir = tempfile.mkdtemp()
            self.cleanups.append(
                partial(shutil.rmtree, tempdir, ignore_errors=True)
            )

            extractall_fn(archive, tempdir)
            self.install(Path(tempdir))


def install_and_restart(code_channel: Optional[Path], packages: Path):
    # skip if there is no code, and if we already done our work
    if code_channel is None or "__SHELL_RELOADED__" in os.environ:
        return

    packages.mkdir(exist_ok=True)

    inst = Installer(packages)
    inst.install(code_channel)
    inst.cleanup()

    python_path = os.environ.get("PYTHONPATH", "")
    python_path = f"{packages}:{python_path}"

    os.environ.update(__SHELL_RELOADED__="1", PYTHONPATH=python_path)

    # restart
    os.execve(
        sys.executable,
        [sys.executable, "-m", "gluonts.shell"] + sys.argv[1:],
        os.environ,
    )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import mxnet as mx

from gluonts.core.component import validated
from gluonts.mx import Tensor
from gluonts.mx.block.decoder import Seq2SeqDecoder
from gluonts.mx.block.enc2dec import Seq2SeqEnc2Dec
from gluonts.mx.block.encoder import Seq2SeqEncoder
from gluonts.mx.block.feature import FeatureEmbedder
from gluonts.mx.block.quantile_output import QuantileOutput
from gluonts.mx.block.scaler import Scaler


class Seq2SeqNetworkBase(mx.gluon.HybridBlock):
    """
    Base network for the :class:`Seq2SeqEstimator`.

    Parameters
    ----------
    scaler : Scaler
        scale of the target time series, both as input or in the output
        distributions
    encoder : encoder
        see encoder.py for possible choices
    enc2dec : encoder to decoder
        see enc2dec.py for possible choices
    decoder : decoder
        see decoder.py for possible choices
    quantile_output : QuantileOutput
        quantile regression output
    kwargs : dict
        a dict of parameters to be passed to the parent initializer
    """

    @validated()
    def __init__(
        self,
        embedder: FeatureEmbedder,
        scaler: Scaler,
        encoder: Seq2SeqEncoder,
        enc2dec: Seq2SeqEnc2Dec,
        decoder: Seq2SeqDecoder,
        quantile_output: QuantileOutput,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.embedder = embedder
        self.scaler = scaler
        self.encoder = encoder
        self.enc2dec = enc2dec
        self.decoder = decoder
        self.quantile_output = quantile_output

        with self.name_scope():
            self.quantile_proj = quantile_output.get_quantile_proj()
            self.loss = quantile_output.get_loss()

    def compute_decoder_outputs(
        self,
        F,
        past_target: Tensor,
        feat_static_cat: Tensor,
        past_feat_dynamic_real: Tensor,
        future_feat_dynamic_real: Tensor,
    ) -> Tensor:
        scaled_target, scale = self.scaler(
            past_target, F.ones_like(past_target)
        )

        embedded_cat = self.embedder(
            feat_static_cat
        )  # (batch_size, num_features * embedding_size)

        encoder_output_static, encoder_output_dynamic = self.encoder(
            scaled_target, embedded_cat, past_feat_dynamic_real
        )
        decoder_input_static, decoder_input_dynamic = self.enc2dec(
            encoder_output_static,
            encoder_output_dynamic,
            future_feat_dynamic_real,
        )
        decoder_output = self.decoder(
            decoder_input_static, decoder_input_dynamic
        )
        scaled_decoder_output = F.broadcast_mul(
            decoder_output, scale.expand_dims(-1).expand_dims(-1)
        )
        return scaled_decoder_output


class Seq2SeqTrainingNetwork(Seq2SeqNetworkBase):
    def hybrid_forward(
        self,
        F,
        past_target: Tensor,
        future_target: Tensor,
        feat_static_cat: Tensor,
        past_feat_dynamic_real: Tensor,
        future_feat_dynamic_real: Tensor,
    ) -> Tensor:
        """

        Parameters
        ----------
        F: mx.symbol or mx.ndarray
            Gluon function space
        past_target: mx.nd.NDArray or mx.sym.Symbol
            past target
        future_target: mx.nd.NDArray or mx.sym.Symbol
            future target
        feat_static_cat: mx.nd.NDArray or mx.sym.Symbol
            static categorical features
        past_feat_dynamic_real: mx.nd.NDArray or mx.sym.Symbol
            past dynamic real-valued features
        future_feat_dynamic_real: mx.nd.NDArray or mx.sym.Symbol
            future dynamic real-valued features

        Returns
        -------
        mx.nd.NDArray or mx.sym.Symbol
           the computed loss
        """
        scaled_decoder_output = self.compute_decoder_outputs(
            F,
            past_target=past_target,
            feat_static_cat=feat_static_cat,
            past_feat_dynamic_real=past_feat_dynamic_real,
            future_feat_dynamic_real=future_feat_dynamic_real,
        )
        projected = self.quantile_proj(scaled_decoder_output)
        loss = self.loss(future_target, projected)
        # TODO: there used to be "nansum" here, to be fully equivalent we
        # TODO: should have a "nanmean" here
        # TODO: shouldn't we sum and divide by the number of observed values
        # TODO: here?
        return loss


class Seq2SeqPredictionNetwork(Seq2SeqNetworkBase):
    def hybrid_forward(
        self,
        F,
        past_target: Tensor,
        feat_static_cat: Tensor,
        past_feat_dynamic_real: Tensor,
        future_feat_dynamic_real: Tensor,
    ) -> Tensor:
        """

        Parameters
        ----------
        F: mx.symbol or mx.ndarray
            Gluon function space
        past_target: mx.nd.NDArray or mx.sym.Symbol
            past target
        feat_static_cat: mx.nd.NDArray or mx.sym.Symbol
            static categorical features
        past_feat_dynamic_real: mx.nd.NDArray or mx.sym.Symbol
            past dynamic real-valued features
        future_feat_dynamic_real: mx.nd.NDArray or mx.sym.Symbol
            future dynamic real-valued features

        Returns
        -------
        mx.nd.NDArray or mx.sym.Symbol
            the predicted sequence
        """
        scaled_decoder_output = self.compute_decoder_outputs(
            F,
            past_target=past_target,
            feat_static_cat=feat_static_cat,
            past_feat_dynamic_real=past_feat_dynamic_real,
            future_feat_dynamic_real=future_feat_dynamic_real,
        )
        predictions = self.quantile_proj(scaled_decoder_output).swapaxes(2, 1)

        return predictions



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Dict, List, Optional, Tuple

import numpy as np

from gluonts.core.component import validated
from gluonts.mx import Tensor
from gluonts.mx.util import cumsum

from .bijection import AffineTransformation, Bijection
from .distribution import Distribution, getF
from .distribution_output import DistributionOutput
from .transformed_distribution import TransformedDistribution


class ISQF(Distribution):
    r"""
    Distribution class for the Incremental (Spline) Quantile Function in the
    paper ``Learning Quantile Functions without Quantile Crossing for
    Distribution-free Time Series Forecasting``
    by Park, Robinson, Aubet, Kan, Gasthaus, Wang

    Parameters
    ----------
    spline_knots, spline_heights
        Tensor parametrizing the x-positions (y-positions) of the spline knots
        Shape: (*batch_shape, (num_qk-1), num_pieces)
    qk_x, qk_y
        Tensor containing the increasing x-positions (y-positions) of the
        quantile knots, Shape: (*batch_shape, num_qk)
    beta_l, beta_r
        Tensor containing the non-negative learnable parameter of the
        left (right) tail, Shape: (*batch_shape,)
    """

    is_reparameterizable = False

    @validated()
    def __init__(
        self,
        spline_knots: Tensor,
        spline_heights: Tensor,
        beta_l: Tensor,
        beta_r: Tensor,
        qk_y: Tensor,
        qk_x: Tensor,
        num_qk: int,
        num_pieces: int,
        tol: float = 1e-4,
    ) -> None:
        self.num_qk, self.num_pieces = num_qk, num_pieces
        self.spline_knots, self.spline_heights = spline_knots, spline_heights
        self.beta_l, self.beta_r = beta_l, beta_r
        self.qk_y_all = qk_y
        self.tol = tol

        F = self.F

        # Get quantile knots (qk) parameters
        (
            self.qk_x,
            self.qk_x_plus,
            self.qk_x_l,
            self.qk_x_r,
        ) = ISQF.parametrize_qk(F, qk_x)
        (
            self.qk_y,
            self.qk_y_plus,
            self.qk_y_l,
            self.qk_y_r,
        ) = ISQF.parametrize_qk(F, qk_y)

        # Get spline knots (sk) parameters
        self.sk_y, self.delta_sk_y = ISQF.parametrize_spline(
            F,
            self.spline_heights,
            self.qk_y,
            self.qk_y_plus,
            self.num_pieces,
            self.tol,
        )
        self.sk_x, self.delta_sk_x = ISQF.parametrize_spline(
            F,
            self.spline_knots,
            self.qk_x,
            self.qk_x_plus,
            self.num_pieces,
            self.tol,
        )

        if self.num_pieces > 1:
            self.sk_x_plus = F.concat(
                F.slice_axis(self.sk_x, axis=-1, begin=1, end=None),
                F.expand_dims(self.qk_x_plus, axis=-1),
                dim=-1,
            )
        else:
            self.sk_x_plus = F.expand_dims(self.qk_x_plus, axis=-1)

        # Get tails parameters
        self.tail_al, self.tail_bl = ISQF.parametrize_tail(
            F, self.beta_l, self.qk_x_l, self.qk_y_l
        )
        self.tail_ar, self.tail_br = ISQF.parametrize_tail(
            F, -self.beta_r, 1 - self.qk_x_r, self.qk_y_r
        )

    @property
    def F(self):
        return getF(self.beta_l)

    @property
    def args(self) -> List:
        return [
            self.spline_knots,
            self.spline_heights,
            self.beta_l,
            self.beta_r,
            self.qk_y_all,
        ]

    @staticmethod
    def parametrize_qk(
        F, quantile_knots: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        r"""
        Function to parametrize the x or y positions
        of the num_qk quantile knots

        Parameters
        ----------
        quantile_knots
            x or y positions of the quantile knots
            shape: (*batch_shape, num_qk)
        Returns
        -------
        qk
            x or y positions of the quantile knots (qk),
            with index=1, ..., num_qk-1,
            shape: (*batch_shape, num_qk-1)
        qk_plus
            x or y positions of the quantile knots (qk),
            with index=2, ..., num_qk,
            shape: (*batch_shape, num_qk-1)
        qk_l
            x or y positions of the left-most quantile knot (qk),
            shape: (*batch_shape)
        qk_r
            x or y positions of the right-most quantile knot (qk),
            shape: (*batch_shape)
        """

        qk = F.slice_axis(quantile_knots, axis=-1, begin=0, end=-1)
        qk_plus = F.slice_axis(quantile_knots, axis=-1, begin=1, end=None)
        qk_l = F.slice_axis(quantile_knots, axis=-1, begin=0, end=1).squeeze(
            axis=-1
        )
        qk_r = F.slice_axis(
            quantile_knots, axis=-1, begin=-1, end=None
        ).squeeze(axis=-1)

        return qk, qk_plus, qk_l, qk_r

    @staticmethod
    def parametrize_spline(
        F,
        spline_knots: Tensor,
        qk: Tensor,
        qk_plus: Tensor,
        num_pieces: int,
        tol: float = 1e-4,
    ) -> Tuple[Tensor, Tensor]:
        r"""
        Function to parametrize the x or y positions of the spline knots
        Parameters
        ----------
        spline_knots
            variable that parameterizes the spline knot positions
        qk
            x or y positions of the quantile knots (qk),
            with index=1, ..., num_qk-1,
            shape: (*batch_shape, num_qk-1)
        qk_plus
            x or y positions of the quantile knots (qk),
            with index=2, ..., num_qk,
            shape: (*batch_shape, num_qk-1)
        num_pieces
            number of spline knot pieces
        tol
            tolerance hyperparameter for numerical stability
        Returns
        -------
        sk
            x or y positions of the spline knots (sk),
            shape: (*batch_shape, num_qk-1, num_pieces)
        delta_sk
            difference of x or y positions of the spline knots (sk),
            shape: (*batch_shape, num_qk-1, num_pieces)
        """

        # The spacing between spline knots is parametrized
        # by softmax function (in [0,1] and sum to 1)
        # We add tol to prevent overflow in computing 1/spacing in spline CRPS
        # After adding tol, it is normalized by
        # (1 + num_pieces * tol) to keep the sum-to-1 property
        delta_x = (F.softmax(spline_knots) + tol) / (1 + num_pieces * tol)

        # TODO: update to mxnet cumsum when it supports axis=-1
        x = cumsum(F, delta_x, exclusive=True)

        qk = F.expand_dims(qk, axis=-1)
        qk_plus = F.expand_dims(qk_plus, axis=-1)

        sk = F.broadcast_add(F.broadcast_mul(x, (qk_plus - qk)), qk)
        delta_sk = F.broadcast_mul(delta_x, (qk_plus - qk))

        return sk, delta_sk

    @staticmethod
    def parametrize_tail(
        F, beta: Tensor, qk_x: Tensor, qk_y: Tensor
    ) -> Tuple[Tensor, Tensor]:
        r"""
        Function to parametrize the tail parameters

        Note that the exponential tails are given by
        q(alpha)
        = a_l log(alpha) + b_l if left tail
        = a_r log(1-alpha) + b_r if right tail

        where
        a_l=1/beta_l, b_l=-a_l*log(qk_x_l)+q(qk_x_l)
        a_r=1/beta_r, b_r=a_r*log(1-qk_x_r)+q(qk_x_r)

        Parameters
        ----------
        beta
            parameterizes the left or right tail, shape: (*batch_shape,)
        qk_x
            left- or right-most x-positions of the quantile knots,
            shape: (*batch_shape,)
        qk_y
            left- or right-most y-positions of the quantile knots,
            shape: (*batch_shape,)
        Returns
        -------
        tail_a
            a_l or a_r as described above
        tail_b
            b_l or b_r as described above
        """

        tail_a = 1 / beta
        tail_b = -tail_a * F.log(qk_x) + qk_y

        return tail_a, tail_b

    def quantile(self, input_alpha: Tensor) -> Tensor:
        return self.quantile_internal(input_alpha, axis=0)

    def quantile_internal(
        self, alpha: Tensor, axis: Optional[int] = None
    ) -> Tensor:
        r"""
        Evaluates the quantile function at the quantile levels input_alpha
        Parameters
        ----------
        alpha
            Tensor of shape = (*batch_shape,) if axis=None, or containing an
            additional axis on the specified position, otherwise
        axis
            Index of the axis containing the different quantile levels which
            are to be computed.
            Read the description below for detailed information
        Returns
        -------
        Tensor
            Quantiles tensor, of the same shape as alpha
        """

        F = self.F
        qk_x, qk_x_l, qk_x_plus = self.qk_x, self.qk_x_l, self.qk_x_plus

        # The following describes the parameters reshaping in
        # quantile_internal, quantile_spline and quantile_tail

        # tail parameters: tail_al, tail_ar, tail_bl, tail_br,
        # shape = (*batch_shape,)
        # spline parameters: sk_x, sk_x_plus, sk_y, sk_y_plus,
        # shape = (*batch_shape, num_qk-1, num_pieces)
        # quantile knots parameters: qk_x, qk_x_plus, qk_y, qk_y_plus,
        # shape = (*batch_shape, num_qk-1)

        # axis=None - passed at inference when num_samples is None
        # shape of input_alpha = (*batch_shape,), will be expanded to
        # (*batch_shape, 1, 1) to perform operation
        # The shapes of parameters are as described above,
        # no reshaping is needed

        # axis=0 - passed at inference when num_samples is not None
        # shape of input_alpha = (num_samples, *batch_shape)
        # it will be expanded to
        # (num_samples, *batch_shape, 1, 1) to perform operation
        #
        # The shapes of tail parameters
        # should be (num_samples, *batch_shape)
        #
        # The shapes of spline parameters
        # should be (num_samples, *batch_shape, num_qk-1, num_pieces)
        #
        # The shapes of quantile knots parameters
        # should be (num_samples, *batch_shape, num_qk-1)
        #
        # We expand axis=0 for all of them

        # axis=-2 - passed at training when we evaluate quantiles at
        # spline knots in order to compute alpha_tilde
        #
        # This is only for the quantile_spline function
        # shape of input_alpha = (*batch_shape, num_qk-1, num_pieces)
        # it will be expanded to
        # (*batch_shape, num_qk-1, num_pieces, 1) to perform operation
        #
        # The shapes of spline and quantile knots parameters should be
        # (*batch_shape, num_qk-1, 1, num_pieces)
        # and (*batch_shape, num_qk-1, 1), respectively
        #
        # We expand axis=-2 and axis=-1 for
        # spline and quantile knots parameters, respectively

        if axis is not None:
            qk_x_l = F.expand_dims(qk_x_l, axis=axis)
            qk_x = F.expand_dims(qk_x, axis=axis)
            qk_x_plus = F.expand_dims(qk_x_plus, axis=axis)

        quantile = F.where(
            F.broadcast_lesser(alpha, qk_x_l),
            self.quantile_tail(alpha, axis=axis, left_tail=True),
            self.quantile_tail(alpha, axis=axis, left_tail=False),
        )

        spline_val = self.quantile_spline(alpha, axis=axis)

        for spline_idx in range(self.num_qk - 1):
            is_in_between = F.broadcast_logical_and(
                F.broadcast_lesser_equal(
                    F.slice_axis(
                        qk_x, axis=-1, begin=spline_idx, end=spline_idx + 1
                    ).squeeze(-1),
                    alpha,
                ),
                F.broadcast_lesser(
                    alpha,
                    F.slice_axis(
                        qk_x_plus,
                        axis=-1,
                        begin=spline_idx,
                        end=spline_idx + 1,
                    ).squeeze(-1),
                ),
            )

            quantile = F.where(
                is_in_between,
                F.slice_axis(
                    spline_val, axis=-1, begin=spline_idx, end=spline_idx + 1
                ).squeeze(-1),
                quantile,
            )

        return quantile

    def quantile_spline(
        self,
        alpha: Tensor,
        axis: Optional[int] = None,
    ) -> Tensor:
        r"""
        Evaluates the spline functions at the
        quantile levels contained in alpha

        Parameters
        ----------
        alpha
            Input quantile levels
        axis
            Axis along which to expand
            For details of input_alpha shape and axis,
            refer to the description in quantile_internal
        Returns
        -------
        Tensor
            Quantiles tensor
            with shape
            = (*batch_shape, num_qk-1) if axis = None
            = (1, *batch_shape, num_qk-1) if axis = 0
            = (*batch_shape, num_qk-1, num_pieces) if axis = -2
        """

        F = self.F
        qk_y = self.qk_y
        sk_x, delta_sk_x, delta_sk_y = (
            self.sk_x,
            self.delta_sk_x,
            self.delta_sk_y,
        )

        if axis is not None:
            qk_y = F.expand_dims(qk_y, axis=0 if axis == 0 else -1)
            sk_x = F.expand_dims(sk_x, axis=axis)
            delta_sk_x = F.expand_dims(delta_sk_x, axis=axis)
            delta_sk_y = F.expand_dims(delta_sk_y, axis=axis)

        if axis is None or axis == 0:
            alpha = F.expand_dims(alpha, axis=-1)

        alpha = F.expand_dims(alpha, axis=-1)

        spline_val = F.broadcast_div(F.broadcast_sub(alpha, sk_x), delta_sk_x)
        spline_val = F.maximum(
            F.minimum(spline_val, F.ones_like(spline_val)),
            F.zeros_like(spline_val),
        )

        return F.broadcast_add(
            qk_y,
            F.sum(
                F.broadcast_mul(spline_val, delta_sk_y),
                axis=-1,
                keepdims=False,
            ),
        )

    def quantile_tail(
        self,
        alpha: Tensor,
        axis: Optional[int] = None,
        left_tail: bool = True,
    ) -> Tensor:
        r"""
        Evaluates the tail functions at the quantile levels contained in alpha
        Parameters
        ----------
        alpha
            Input quantile levels
        axis
            Axis along which to expand
            For details of input_alpha shape and axis,
            refer to the description in quantile_internal
        left_tail
            If True, compute the quantile for the left tail
            Otherwise, compute the quantile for the right tail
        Returns
        -------
        Tensor
            Quantiles tensor, of the same shape as alpha
        """

        F = self.F

        if left_tail:
            tail_a, tail_b = self.tail_al, self.tail_bl
        else:
            tail_a, tail_b = self.tail_ar, self.tail_br
            alpha = 1 - alpha

        if axis is not None:
            tail_a, tail_b = (
                F.expand_dims(tail_a, axis=axis),
                F.expand_dims(tail_b, axis=axis),
            )

        return F.broadcast_add(F.broadcast_mul(tail_a, F.log(alpha)), tail_b)

    def cdf_spline(self, z: Tensor) -> Tensor:
        r"""
        For observations z and splines defined in [qk_x[k], qk_x[k+1]]
        Computes the quantile level alpha_tilde such that
        alpha_tilde
        = q^{-1}(z) if z is in-between qk_x[k] and qk_x[k+1]
        = qk_x[k] if z<qk_x[k]
        = qk_x[k+1] if z>qk_x[k+1]
        Parameters
        ----------
        z
            Observation, shape = (*batch_shape,)
        Returns
        -------
        alpha_tilde
            Corresponding quantile level, shape = (*batch_shape, num_qk-1)
        """

        F = self.F

        qk_y, qk_y_plus = self.qk_y, self.qk_y_plus
        qk_x, qk_x_plus = self.qk_x, self.qk_x_plus
        sk_x, delta_sk_x, delta_sk_y = (
            self.sk_x,
            self.delta_sk_x,
            self.delta_sk_y,
        )

        z_expand = F.expand_dims(z, axis=-1)

        if self.num_pieces > 1:
            qk_y_expand = F.expand_dims(qk_y, axis=-1)
            z_expand_twice = F.expand_dims(z_expand, axis=-1)

            knots_eval = self.quantile_spline(sk_x, axis=-2)

            # Compute \sum_{s=0}^{s_0-1} \Delta sk_y[s],
            # where \Delta sk_y[s] = (sk_y[s+1]-sk_y[s])
            mask_sum_s0 = F.broadcast_lesser(knots_eval, z_expand_twice)
            mask_sum_s0_minus = F.concat(
                F.slice_axis(mask_sum_s0, axis=-1, begin=1, end=None),
                F.zeros_like(qk_y_expand),
                dim=-1,
            )
            sum_delta_sk_y = F.sum(
                F.broadcast_mul(mask_sum_s0_minus, delta_sk_y),
                axis=-1,
                keepdims=False,
            )

            mask_s0_only = mask_sum_s0 - mask_sum_s0_minus
            # Compute (sk_x[s_0+1]-sk_x[s_0])/(sk_y[s_0+1]-sk_y[s_0])
            frac_s0 = F.sum(
                (mask_s0_only * delta_sk_x) / delta_sk_y,
                axis=-1,
                keepdims=False,
            )

            # Compute sk_x_{s_0}
            sk_x_s0 = F.sum(mask_s0_only * sk_x, axis=-1, keepdims=False)

            # Compute alpha_tilde
            alpha_tilde = (
                sk_x_s0
                + (F.broadcast_sub(z_expand, qk_y) - sum_delta_sk_y) * frac_s0
            )

        else:
            # num_pieces=1, ISQF reduces to IQF
            alpha_tilde = qk_x + F.broadcast_sub(z_expand, qk_y) / (
                qk_y_plus - qk_y
            ) * (qk_x_plus - qk_x)

        alpha_tilde = F.broadcast_minimum(
            F.broadcast_maximum(alpha_tilde, qk_x), qk_x_plus
        )

        return alpha_tilde

    def cdf_tail(self, z: Tensor, left_tail: bool = True) -> Tensor:
        r"""
        Computes the quantile level alpha_tilde such that
        alpha_tilde
        = q^{-1}(z) if z is in the tail region
        = qk_x_l or qk_x_r if z is in the non-tail region
        Parameters
        ----------
        z
            Observation, shape = (*batch_shape,)
        left_tail
            If True, compute alpha_tilde for the left tail
            Otherwise, compute alpha_tilde for the right tail
        Returns
        -------
        alpha_tilde
            Corresponding quantile level, shape = (*batch_shape,)
        """

        F = self.F

        if left_tail:
            tail_a, tail_b, qk_x = self.tail_al, self.tail_bl, self.qk_x_l
        else:
            tail_a, tail_b, qk_x = self.tail_ar, self.tail_br, 1 - self.qk_x_r

        log_alpha_tilde = F.minimum((z - tail_b) / tail_a, F.log(qk_x))
        alpha_tilde = F.exp(log_alpha_tilde)
        return alpha_tilde if left_tail else 1 - alpha_tilde

    def crps_tail(self, z: Tensor, left_tail: bool = True) -> Tensor:
        r"""
        Compute CRPS in analytical form for left/right tails
        Parameters
        ----------
        z
            Observation to evaluate. shape = (*batch_shape,)
        left_tail
            If True, compute CRPS for the left tail
            Otherwise, compute CRPS for the right tail
        Returns
        -------
        Tensor
            Tensor containing the CRPS, of the same shape as z
        """

        F = self.F
        alpha_tilde = self.cdf_tail(z, left_tail=left_tail)

        if left_tail:
            tail_a, tail_b, qk_x, qk_y = (
                self.tail_al,
                self.tail_bl,
                self.qk_x_l,
                self.qk_y_l,
            )
            term1 = (z - tail_b) * (qk_x**2 - 2 * qk_x + 2 * alpha_tilde)
            term2 = qk_x**2 * tail_a * (-F.log(qk_x) + 0.5)
            term2 = term2 + 2 * F.where(
                z < qk_y,
                qk_x * tail_a * (F.log(qk_x) - 1)
                + alpha_tilde * (-z + tail_b + tail_a),
                F.zeros_like(qk_x),
            )
        else:
            tail_a, tail_b, qk_x, qk_y = (
                self.tail_ar,
                self.tail_br,
                self.qk_x_r,
                self.qk_y_r,
            )
            term1 = (z - tail_b) * (-1 - qk_x**2 + 2 * alpha_tilde)
            term2 = tail_a * (
                -0.5 * (qk_x + 1) ** 2
                + (qk_x**2 - 1) * F.log(1 - qk_x)
                + 2 * alpha_tilde
            )
            term2 = term2 + 2 * F.where(
                z > qk_y,
                (1 - alpha_tilde) * (z - tail_b),
                tail_a * (1 - qk_x) * F.log(1 - qk_x),
            )

        return term1 + term2

    def crps_spline(self, z: Tensor) -> Tensor:
        r"""
        Compute CRPS in analytical form for the spline
        Parameters
        ----------
        z
            Observation to evaluate. shape = (*batch_shape,)
        Returns
        -------
        Tensor
            Tensor containing the CRPS, of the same shape as z
        """

        F = self.F
        qk_x, qk_x_plus, qk_y = self.qk_x, self.qk_x_plus, self.qk_y
        sk_x, sk_x_plus = self.sk_x, self.sk_x_plus
        delta_sk_x, delta_sk_y = self.delta_sk_x, self.delta_sk_y

        z_expand, qk_x_plus_expand = (
            F.expand_dims(z, axis=-1),
            F.expand_dims(qk_x_plus, axis=-1),
        )

        alpha_tilde = self.cdf_spline(z)
        alpha_tilde_expand = F.expand_dims(alpha_tilde, axis=-1)

        r = F.broadcast_minimum(
            F.broadcast_maximum(alpha_tilde_expand, sk_x), sk_x_plus
        )

        coeff1 = (
            -2 / 3 * sk_x_plus**3
            + sk_x * sk_x_plus**2
            + sk_x_plus**2
            - (1 / 3) * sk_x**3
            - 2 * sk_x * sk_x_plus
            - r**2
            + 2 * sk_x * r
        )

        coeff2 = F.broadcast_add(
            -2 * F.broadcast_maximum(alpha_tilde_expand, sk_x_plus)
            + sk_x_plus**2,
            2 * qk_x_plus_expand - qk_x_plus_expand**2,
        )

        result = (
            (qk_x_plus**2 - qk_x**2) * F.broadcast_sub(z_expand, qk_y)
            + 2
            * F.broadcast_sub(qk_x_plus, alpha_tilde)
            * F.broadcast_sub(qk_y, z_expand)
            + F.sum(
                (delta_sk_y / delta_sk_x) * coeff1, axis=-1, keepdims=False
            )
            + F.sum(delta_sk_y * coeff2, axis=-1, keepdims=False)
        )

        return F.sum(result, axis=-1, keepdims=False)

    def loss(self, z: Tensor) -> Tensor:
        return self.crps(z)

    def crps(self, z: Tensor) -> Tensor:
        r"""
        Compute CRPS in analytical form
        Parameters
        ----------
        z
            Observation to evaluate. Shape = (*batch_shape,)
        Returns
        -------
        Tensor
            Tensor containing the CRPS, of the same shape as z
        """

        crps_lt = self.crps_tail(z, left_tail=True)
        crps_rt = self.crps_tail(z, left_tail=False)

        return crps_lt + crps_rt + self.crps_spline(z)

    def cdf(self, z: Tensor) -> Tensor:
        r"""
        Computes the quantile level alpha_tilde such that
        q(alpha_tilde) = z
        Parameters
        ----------
        z
            Tensor of shape = (*batch_shape,)
        Returns
        -------
        alpha_tilde
            Tensor of shape = (*batch_shape,)
        """

        F = self.F
        qk_y, qk_y_l, qk_y_plus = self.qk_y, self.qk_y_l, self.qk_y_plus

        alpha_tilde = F.where(
            z < qk_y_l,
            self.cdf_tail(z, left_tail=True),
            self.cdf_tail(z, left_tail=False),
        )

        spline_alpha_tilde = self.cdf_spline(z)

        for i in range(self.num_qk - 1):
            is_in_between = F.broadcast_logical_and(
                F.slice_axis(qk_y, axis=-1, begin=i, end=i + 1).squeeze(-1)
                <= z,
                z
                < F.slice_axis(qk_y_plus, axis=-1, begin=i, end=i + 1).squeeze(
                    -1
                ),
            )

            alpha_tilde = F.where(
                is_in_between,
                F.slice_axis(
                    spline_alpha_tilde, axis=-1, begin=i, end=i + 1
                ).squeeze(-1),
                alpha_tilde,
            )

        return alpha_tilde

    def sample(
        self, num_samples: Optional[int] = None, dtype=np.float32
    ) -> Tensor:
        r"""
        Function used to draw random samples
        Parameters
        ----------
        num_samples
            number of samples
        dtype
            data type
        Returns
        -------
        Tensor
            Tensor of shape (*batch_shape,) if num_samples = None
            else (num_samples, *batch_shape)
        """

        F = self.F

        # if num_samples=None then input_alpha should have the same shape
        # as beta_l, i.e., (*batch_shape,)
        # else u should be (num_samples, *batch_shape)
        alpha = F.random.uniform_like(
            data=(
                self.beta_l
                if num_samples is None
                else self.beta_l.expand_dims(axis=0).repeat(
                    axis=0, repeats=num_samples
                )
            )
        )

        sample = self.quantile(alpha)

        if num_samples is None:
            sample = F.squeeze(sample, axis=0)

        return sample

    @property
    def batch_shape(self) -> Tuple:
        return self.beta_l.shape

    @property
    def event_shape(self) -> Tuple:
        return ()

    @property
    def event_dim(self) -> int:
        return 0


class ISQFOutput(DistributionOutput):
    r"""
    DistributionOutput class for the Incremental (Spline) Quantile Function

    Parameters
    ----------
    num_pieces
        number of spline pieces for each spline
        ISQF reduces to IQF when num_pieces = 1
    alpha
        Tensor containing the x-positions of quantile knots
    tol
        tolerance for numerical safeguarding
    """

    distr_cls: type = ISQF

    @validated()
    def __init__(
        self, num_pieces: int, qk_x: List[float], tol: float = 1e-4
    ) -> None:
        # ISQF reduces to IQF when num_pieces = 1

        super().__init__(self)

        assert (
            isinstance(num_pieces, int) and num_pieces > 0
        ), "num_pieces should be an integer and greater than 0"

        self.num_pieces = num_pieces
        self.qk_x = sorted(qk_x)
        self.num_qk = len(qk_x)
        self.tol = tol
        self.args_dim: Dict[str, int] = {
            "spline_knots": (self.num_qk - 1) * num_pieces,
            "spline_heights": (self.num_qk - 1) * num_pieces,
            "beta_l": 1,
            "beta_r": 1,
            "quantile_knots": self.num_qk,
        }

    @classmethod
    def domain_map(
        cls,
        F,
        *args: Tensor,
        tol: float = 1e-4,
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor]:
        """
        Domain map function The inputs of this function are specified by
        self.args_dim knots, heights:

        parameterizing the x-/ y-positions of the spline knots,
        shape = (*batch_shape, (num_qk-1)*num_pieces)
        q:
        parameterizing the y-positions of the quantile knots,
        shape = (*batch_shape, num_qk)
        beta_l, beta_r:
        parameterizing the left/right tail, shape = (*batch_shape, 1)
        """

        try:
            spline_knots, spline_heights, beta_l, beta_r, quantile_knots = args
        except ValueError:
            raise ValueError(
                "Failed to unpack args of domain_map. Double check your input."
            )

        # Add tol to prevent the y-distance of
        # two quantile knots from being too small
        #
        # Because in this case the spline knots could be squeezed together
        # and cause overflow in spline CRPS computation
        qk_y = F.concat(
            F.slice_axis(quantile_knots, axis=-1, begin=0, end=1),
            F.abs(F.slice_axis(quantile_knots, axis=-1, begin=1, end=None))
            + tol,
            dim=-1,
        )
        # TODO: update to mxnet cumsum when it supports axis=-1
        qk_y = cumsum(F, qk_y)

        # Prevent overflow when we compute 1/beta
        beta_l, beta_r = (
            F.abs(beta_l.squeeze(axis=-1)) + tol,
            F.abs(beta_r.squeeze(axis=-1)) + tol,
        )

        return spline_knots, spline_heights, beta_l, beta_r, qk_y

    def distribution(
        self,
        distr_args,
        loc: Optional[Tensor] = None,
        scale: Optional[Tensor] = None,
    ) -> ISQF:
        """
        function outputing the distribution class
        distr_args: distribution arguments
        loc: shift to the data mean
        scale: scale to the data
        """

        distr_args, qk_x = self.reshape_spline_args(distr_args, self.qk_x)

        if scale is None:
            return self.distr_cls(
                *distr_args, qk_x, self.num_qk, self.num_pieces, self.tol
            )
        else:
            distr = self.distr_cls(
                *distr_args, qk_x, self.num_qk, self.num_pieces, self.tol
            )
            return TransformedISQF(
                distr, [AffineTransformation(loc=loc, scale=scale)]
            )

    def reshape_spline_args(self, distr_args, qk_x):
        """
        auxiliary function reshaping knots and heights to (*batch_shape,
        num_qk-1, num_pieces) alpha to (*batch_shape, num_qk)
        """

        spline_knots, spline_heights = distr_args[0], distr_args[1]
        beta_l = distr_args[2]
        qk_y = distr_args[4]
        F = getF(beta_l)

        # FIXME number 1
        # Convert alpha from list of len=num_qk to
        # Tensor of shape (*batch_shape, num_qk)
        #
        # For example, if alpha = [0.1, 0.5, 0.9],
        # then alpha_reshape will be a Tensor of shape (*batch_shape, 3)
        # with the last dimension being [0.1, 0.5, 0.9]
        #
        # In PyTorch, it would be torch.tensor(alpha).repeat(*batch_shape,1)
        qk_x_reshape = F.concat(
            *[
                F.expand_dims(F.ones_like(beta_l), axis=-1) * qk_x[i]
                for i in range(self.num_qk)
            ],
            dim=-1,
        )

        # FIXME number 2
        # knots and heights have shape (*batch_shape, (num_qk-1)*num_pieces)
        # I want to convert the shape to (*batch_shape, (num_qk-1), num_pieces)
        # Here I make a shape_holder with target_shape, and use reshape_like

        # create a shape holder of shape (*batch_shape, num_qk-1, num_pieces)
        shape_holder = F.repeat(
            F.expand_dims(
                F.slice_axis(qk_y, axis=-1, begin=0, end=-1), axis=-1
            ),
            repeats=self.num_pieces,
            axis=-1,
        )

        spline_knots_reshape = F.reshape_like(spline_knots, shape_holder)
        spline_heights_reshape = F.reshape_like(spline_heights, shape_holder)
        distr_args_reshape = (
            spline_knots_reshape,
            spline_heights_reshape,
        ) + distr_args[2:]

        return distr_args_reshape, qk_x_reshape

    @property
    def event_shape(self) -> Tuple:
        return ()


class TransformedISQF(TransformedDistribution, ISQF):
    # DistributionOutput class for the case when loc/scale is not None
    @validated()
    def __init__(
        self, base_distribution: ISQF, transforms: List[Bijection]
    ) -> None:
        super().__init__(base_distribution, transforms)

    def crps(self, y: Tensor) -> Tensor:
        F = getF(y)
        z = y
        scale = 1.0
        for t in self.transforms[::-1]:
            assert isinstance(
                t, AffineTransformation
            ), "Not an AffineTransformation"
            z = t.f_inv(z)
            scale *= t.scale
        p = self.base_distribution.crps(z)
        return F.broadcast_mul(p, scale)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# Standard library imports
from functools import partial
from typing import Callable

# Third-party imports
import mxnet as mx
import mxnet.gluon.nn as nn

# First-party imports
from gluonts.core.component import validated
from gluonts.mx import Tensor


def get_activation(activation: str, **kwargs) -> nn.HybridBlock:
    """

    Parameters
    ----------
    activation
        Activation type

    Returns
    -------
    mxnet.gluon.HybridBlock
        Activation object

    """
    if activation in ["relu", "sigmoid", "softrelu", "softsign", "tanh"]:
        return nn.Activation(activation=activation, **kwargs)
    if activation == "lrelu":
        return nn.LeakyReLU(alpha=0.2, **kwargs)
    if activation == "elu":
        return nn.ELU(**kwargs)
    if activation == "swish":
        return nn.Swish(**kwargs)
    if activation == "lipswish":
        return LipSwish(**kwargs)
    raise NotImplementedError(activation)


def get_activation_deriv(act: nn.HybridBlock) -> Callable:
    """

    Parameters
    ----------
    act
        Activation object
    Returns
    -------
    Callable
        Derivative function of associated activation

    """
    if isinstance(act, nn.Activation):
        activation = act._act_type
        if activation == "relu":
            raise NotImplementedError(activation)
        if activation == "sigmoid":
            raise NotImplementedError(activation)
        if activation == "tanh":
            return deriv_tanh
        if activation == "softrelu":
            return deriv_softrelu
        if activation == "softsign":
            raise NotImplementedError(activation)
    if isinstance(act, nn.ELU):
        return partial(deriv_elu, alpha=act._alpha)
    if isinstance(act, nn.Swish):
        return partial(deriv_swish, beta=act._beta)
    if isinstance(act, LipSwish):
        return partial(deriv_lipswish, beta=act.params.get("beta").data())
    raise NotImplementedError(
        f'No derivative function for activation "{act.__class__.__name__}"'
    )


def deriv_tanh(F, x: Tensor) -> Tensor:
    """
    Derivative function of Tanh activation computed at point `x`.

    Parameters
    ----------
    F
        A module that can either refer to the Symbol API or the NDArray API in
        MXNet.

    x
        Input tensor

    Returns
    -------
    Tensor
        Derivative tensor
    """
    return 1 - F.tanh(x) ** 2


def deriv_softrelu(F, x: Tensor) -> Tensor:
    """
    Derivative function of SoftRelu activation computed at point `x`.

    Parameters
    ----------
    F
        A module that can either refer to the Symbol API or the NDArray API in
        MXNet.
    x
        Input tensor

    Returns
    -------
    Tensor
        Derivative tensor
    """
    e = mx.nd.exp(x)
    return e / (1 + e)


def deriv_elu(F, x: Tensor, alpha: float = 1.0) -> Tensor:
    """
    Derivative function of Elu activation computed at point `x`.

    Parameters
    ----------
    F
        A module that can either refer to the Symbol API or the NDArray API in
        MXNet.
    x
        Input tensor
    alpha
        alpha parameter of Elu

    Returns
    -------
    Tensor
        Derivative tensor
    """
    m = x > 0
    return m + (1 - m) * (F.LeakyReLU(x, act_type="elu", slope=alpha) + alpha)


def deriv_swish(F, x: Tensor, beta: Tensor) -> Tensor:
    """
    Derivative function of Swish activation computed at point `x`.

    Parameters
    ----------
    F
        A module that can either refer to the Symbol API or the NDArray API in
        MXNet.
    x
        Input tensor
    beta
        beta parameter of LipSwish

    Returns
    -------
    Tensor
        Derivative tensor
    """
    f = x * F.sigmoid(beta * x, name="fwd")
    return beta * f + F.sigmoid(beta * x) * (1 - beta * f)


def deriv_lipswish(F, x: Tensor, beta: Tensor) -> Tensor:
    """
    Derivative function of LipSwish activation computed at point `x`.
    Parameters
    ----------
    F
        A module that can either refer to the Symbol API or the NDArray API in
        MXNet.
    x
        Input tensor
    beta
        beta parameter in LipSwish activation
    Returns
    -------
    Tensor
        Derivative tensor

    """
    return deriv_swish(F, x, beta) / 1.1


class LipSwish(nn.HybridBlock):
    """
    Implemented LipSwish activation, i.e. LipSwish(z) := Swish(z)/ 1.1 with a
    learnable parameter beta.
    """

    @validated()
    def __init__(
        self,
        beta_initializer: mx.init.Initializer = mx.init.Constant(1.0),
        **kwargs,
    ):
        super().__init__(**kwargs)
        with self.name_scope():
            self.beta = self.params.get(
                "beta", shape=(1,), init=beta_initializer
            )

    def hybrid_forward(self, F, x: Tensor, beta: Tensor) -> Tensor:
        """

        Parameters
        ----------
        F
            A module that can either refer to the Symbol API or the NDArray
            API in MXNet.
        x
            Input tensor
        beta
            beta parameter of activation

        Returns
        -------
        Tensor
            output of forward

        """
        return x * F.sigmoid(beta * x, name="fwd") / 1.1



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

__all__ = [
    "PyTorchLightningEstimator",
    "PyTorchPredictor",
    "DeepNPTSEstimator",
    "DeepAREstimator",
    "SimpleFeedForwardEstimator",
    "TemporalFusionTransformerEstimator",
]
import torch
from .model.estimator import PyTorchLightningEstimator
from .model.predictor import PyTorchPredictor
from .model.deep_npts import DeepNPTSEstimator
from .model.deepar import DeepAREstimator
from .model.simple_feedforward import SimpleFeedForwardEstimator
from .model.tft import TemporalFusionTransformerEstimator


from . import prelude as _  # noqa


if torch.cuda.is_available():
    torch.multiprocessing.set_start_method("spawn", force=True)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations

import dataclasses
from typing import Any, Optional, NamedTuple, Union
from typing_extensions import Literal


from gluonts import maybe

from ._period import Periods, Period
from ._util import _replace


LeftOrRight = Literal["l", "r"]


class Pad(NamedTuple):
    """Indicator for padded values.

    >>> from gluonts.zebras import time_series
    >>> ts = time_series([1, 2, 3]).pad(0, left=2, right=2)
    >>> assert len(ts == 7)
    >>> assert list(ts) == [0, 0, 1, 2, 3, 0, 0]
    >>> assert ts._pad.left == 2
    >>> assert ts._pad.right == 2

    """

    left: int = 0
    right: int = 0

    def extend(self, left, right):
        return Pad(
            left=max(0, self.left + left),
            right=max(0, self.right + right),
        )


@dataclasses.dataclass
class ILoc:
    tb: TimeBase

    def __getitem__(self, idx):
        return self.tb._slice_tdim(idx)


@dataclasses.dataclass
class TimeView:
    tf: TimeBase

    def __getitem__(self, idx):
        assert isinstance(idx, slice)
        assert idx.step is None or idx.step == 1

        start = maybe.map(idx.start, self.tf.index_of)
        stop = maybe.map(idx.stop, self.tf.index_of)

        return ILoc(self.tf)[start:stop]


class TimeBase:
    index: Any

    def _slice_tdim(self, idx):
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def pad(self, value, left: int = 0, right: int = 0) -> TimeBase:
        raise NotImplementedError

    @property
    def iloc(self):
        return ILoc(self)

    @property
    def loc(self):
        return TimeView(self)

    @property
    def start(self):
        return self.iloc[0]

    @property
    def end(self):
        return self.iloc[-1]

    def head(self, count: int) -> Periods:
        return self.iloc[:count]

    def tail(self, count: int) -> Periods:
        if count is None:
            return self

        return self.iloc[-count:]

    def resize(
        self,
        length: Optional[int],
        pad_value=0,
        pad: LeftOrRight = "l",
        skip: LeftOrRight = "r",
    ) -> TimeBase:
        """Force time frame to have length ``length``.

        This pads or slices the time frame, depending on whether its size is
        smaller or bigger than the required length.

        By default we pad values on the left, and skip on the right.
        """
        assert pad in ("l", "r")
        assert skip in ("l", "r")

        if length is None or len(self) == length:
            return self

        if len(self) < length:
            left = right = 0
            if pad == "l":
                left = length - len(self)
            else:
                right = length - len(self)

            return self.pad(pad_value, left=left, right=right)

        if skip == "l":
            return self.iloc[len(self) - length :]
        else:
            return self.iloc[: length - len(self)]

    def index_of(self, period: Union[Period, str]) -> int:
        assert self.index is not None

        return self.index.index_of(period)

    def __getitem__(self, idx: Union[slice, int, str]):
        subtype: Optional[int]

        if isinstance(idx, int):
            subtype = idx
        else:
            assert isinstance(idx, slice)

            start = idx.start
            if start is not None or not isinstance(start, int):
                if isinstance(start, (Period, str)):
                    start = self.index_of(start)

            stop = idx.stop
            if stop is not None or not isinstance(stop, int):
                if isinstance(stop, (Period, str)):
                    stop = self.index_of(stop)

            idx = slice(start, stop, idx.step)

        return self.iloc[idx]

        raise RuntimeError(f"Unsupported type {subtype}.")

    def with_index(self, index):
        return _replace(self, index=index)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Tuple

from mxnet.gluon import nn

from gluonts.core.component import validated
from gluonts.mx import Tensor
from gluonts.mx.block.cnn import CausalConv1D
from gluonts.mx.block.mlp import MLP
from gluonts.mx.block.rnn import RNN


class Seq2SeqEncoder(nn.HybridBlock):
    """
    Abstract class for the encoder.

    An encoder takes a `target` sequence with corresponding covariates and maps
    it into a static latent and a dynamic latent code with the same length as
    the `target` sequence.
    """

    def hybrid_forward(
        self,
        F,
        target: Tensor,
        static_features: Tensor,
        dynamic_features: Tensor,
    ) -> Tuple[Tensor, Tensor]:
        """
        Parameters
        ----------
        F:
            A module that can either refer to the Symbol API or the NDArray
            API in MXNet.
        target
            target time series,
            shape (batch_size, sequence_length)
        static_features
            static features,
            shape (batch_size, num_feat_static)
        dynamic_features
            dynamic_features,
            shape (batch_size, sequence_length, num_feat_dynamic)

        Returns
        -------
        Tensor
            static code,
            shape (batch_size, num_feat_static)
        Tensor
            dynamic code,
            shape (batch_size, sequence_length, num_feat_dynamic)
        """
        raise NotImplementedError

    def _assemble_inputs(
        self,
        F,
        target: Tensor,
        static_features: Tensor,
        dynamic_features: Tensor,
    ) -> Tensor:
        """
        Assemble features from target, static features, and the dynamic
        features.

        Parameters
        ----------
        F
            A module that can either refer to the Symbol API or the NDArray
            API in MXNet.
        target
            target time series,
            shape (batch_size, sequence_length, 1)
        static_features
            static features,
            shape (batch_size, num_feat_static)
        dynamic_features
            dynamic_features,
            shape (batch_size, sequence_length, num_feat_dynamic)

        Returns
        -------
        Tensor
            combined features,
            shape (batch_size, sequence_length,
                   num_feat_static + num_feat_dynamic + 1)
        """

        helper_ones = F.ones_like(target)  # Ones of (N, T, 1)
        tiled_static_features = F.batch_dot(
            helper_ones, static_features.expand_dims(1)
        )  # (N, T, C)
        inputs = F.concat(
            target, tiled_static_features, dynamic_features, dim=2
        )  # (N, T, C)
        return inputs


# TODO: fix handling of static features
class HierarchicalCausalConv1DEncoder(Seq2SeqEncoder):
    """
    Defines a stack of dilated convolutions as the encoder.

    See the following paper for details:
    1. Van Den Oord, A., Dieleman, S., Zen, H., Simonyan, K., Vinyals, O.,
    Graves, A., Kalchbrenner, N., Senior, A.W. and Kavukcuoglu, K., 2016,
    September. WaveNet: A generative model for raw audio. In SSW (p. 125).

    Parameters
    ----------
    dilation_seq
        dilation for each convolution in the stack.
    kernel_size_seq
        kernel size for each convolution in the stack.
    channels_seq
        number of channels for each convolution in the stack.
    use_residual
        flag to toggle using residual connections.
    use_static_feat
        flag to toggle whether to use use_static_feat as input to the encoder
    use_dynamic_feat
        flag to toggle whether to use use_dynamic_feat as input to the encoder
    """

    @validated()
    def __init__(
        self,
        dilation_seq: List[int],
        kernel_size_seq: List[int],
        channels_seq: List[int],
        use_residual: bool = False,
        use_static_feat: bool = False,
        use_dynamic_feat: bool = False,
        **kwargs,
    ) -> None:
        assert all(
            [x > 0 for x in dilation_seq]
        ), "`dilation_seq` values must be greater than zero"
        assert all(
            [x > 0 for x in kernel_size_seq]
        ), "`kernel_size_seq` values must be greater than zero"
        assert all(
            [x > 0 for x in channels_seq]
        ), "`channel_dim_seq` values must be greater than zero"

        super().__init__(**kwargs)

        self.use_residual = use_residual
        self.use_static_feat = use_static_feat
        self.use_dynamic_feat = use_dynamic_feat
        self.cnn = nn.HybridSequential()

        it = zip(channels_seq, kernel_size_seq, dilation_seq)
        for layer_no, (channels, kernel_size, dilation) in enumerate(it):
            convolution = CausalConv1D(
                channels=channels,
                kernel_size=kernel_size,
                dilation=dilation,
                activation="relu",
                prefix=f"conv_{layer_no:#02d}'_",
            )
            self.cnn.add(convolution)

    def hybrid_forward(
        self,
        F,
        target: Tensor,
        static_features: Tensor,
        dynamic_features: Tensor,
    ) -> Tuple[Tensor, Tensor]:
        """
        Parameters
        ----------
        F
            A module that can either refer to the Symbol API or the NDArray
            API in MXNet.
        target
            target time series,
            shape (batch_size, sequence_length, 1)
        static_features
            static features,
            shape (batch_size, num_feat_static)
        dynamic_features
            dynamic_features,
            shape (batch_size, sequence_length, num_feat_dynamic)
        Returns
        -------
        Tensor
            static code,
            shape (batch_size, channel_seqs + (1) if use_residual)
        Tensor
            dynamic code,
            shape (batch_size, sequence_length, channel_seqs + (1) if
            use_residual)
        """

        if self.use_dynamic_feat and self.use_static_feat:
            inputs = self._assemble_inputs(
                F,
                target=target,
                static_features=static_features,
                dynamic_features=dynamic_features,
            )
        elif self.use_dynamic_feat:
            inputs = F.concat(target, dynamic_features, dim=2)  # (N, T, C)
        else:
            # For now, static features only used when dynamic feat enabled
            inputs = target

        # NTC -> NCT (or NCW)
        ct = inputs.swapaxes(1, 2)
        ct = self.cnn(ct)
        ct = ct.swapaxes(1, 2)

        # now we are back in NTC
        if self.use_residual:
            ct = F.concat(ct, target, dim=2)

        # return the last state as the static code
        static_code = F.slice_axis(ct, axis=1, begin=-1, end=None)
        static_code = F.squeeze(static_code, axis=1)

        return static_code, ct


class RNNEncoder(Seq2SeqEncoder):
    """
    Defines RNN encoder that uses covariates and target as input to the RNN if
    desired.

    Parameters
    ----------
    mode
        type of the RNN. Can be either: rnn_relu (RNN with relu activation),
        rnn_tanh, (RNN with tanh activation), lstm or gru.
    hidden_size
        number of units per hidden layer.
    num_layers
        number of hidden layers.
    bidirectional
        toggle use of bi-directional RNN as encoder.
    use_static_feat
        flag to toggle whether to use use_static_feat as input to the encoder
    use_dynamic_feat
        flag to toggle whether to use use_dynamic_feat as input to the encoder
    """

    @validated()
    def __init__(
        self,
        mode: str,
        hidden_size: int,
        num_layers: int,
        bidirectional: bool,
        use_static_feat: bool = False,
        use_dynamic_feat: bool = False,
        **kwargs,
    ) -> None:
        assert num_layers > 0, "`num_layers` value must be greater than zero"
        assert hidden_size > 0, "`hidden_size` value must be greater than zero"

        super().__init__(**kwargs)

        self.mode = mode
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bidirectional = bidirectional
        self.use_static_feat = use_static_feat
        self.use_dynamic_feat = use_dynamic_feat

        with self.name_scope():
            self.rnn = RNN(mode, hidden_size, num_layers, bidirectional)

    def hybrid_forward(
        self,
        F,
        target: Tensor,
        static_features: Tensor,
        dynamic_features: Tensor,
    ) -> Tuple[Tensor, Tensor]:
        """
        Parameters
        ----------
        F
            A module that can either refer to the Symbol API or the NDArray
            API in MXNet.
        target
            target time series,
            shape (batch_size, sequence_length, 1)
        static_features
            static features,
            shape (batch_size, num_feat_static)
        dynamic_features
            dynamic_features,
            shape (batch_size, sequence_length, num_feat_dynamic)

        Returns
        -------
        Tensor
            static code,
            shape (batch_size, num_feat_static)
        Tensor
            dynamic code,
            shape (batch_size, sequence_length, num_feat_dynamic)
        """
        if self.use_dynamic_feat and self.use_static_feat:
            inputs = self._assemble_inputs(
                F,
                target=target,
                static_features=static_features,
                dynamic_features=dynamic_features,
            )
        elif self.use_dynamic_feat:
            inputs = F.concat(target, dynamic_features, dim=2)  # (N, T, C)
        else:
            inputs = target

        dynamic_code = self.rnn(inputs)
        static_code = F.slice_axis(dynamic_code, axis=1, begin=-1, end=None)
        return static_code, dynamic_code


class MLPEncoder(Seq2SeqEncoder):
    """
    Defines a multilayer perceptron used as an encoder.

    Parameters
    ----------
    layer_sizes
        number of hidden units per layer.
    kwargs
    """

    @validated()
    def __init__(self, layer_sizes: List[int], **kwargs) -> None:
        super().__init__(**kwargs)
        self.model = MLP(layer_sizes, flatten=True)

    def hybrid_forward(
        self,
        F,
        target: Tensor,
        static_features: Tensor,
        dynamic_features: Tensor,
    ) -> Tuple[Tensor, Tensor]:
        """
        Parameters
        ----------
        F
            A module that can either refer to the Symbol API or the NDArray
            API in MXNet.
        target
            target time series,
            shape (batch_size, sequence_length)
        static_features
            static features,
            shape (batch_size, num_feat_static)
        dynamic_features
            dynamic_features,
            shape (batch_size, sequence_length, num_feat_dynamic)

        Returns
        -------
        Tensor
            static code,
            shape (batch_size, num_feat_static)
        Tensor
            dynamic code,
            shape (batch_size, sequence_length, num_feat_dynamic)
        """

        inputs = self._assemble_inputs(
            F, target, static_features, dynamic_features
        )
        static_code = self.model(inputs)
        dynamic_code = F.zeros_like(target).expand_dims(2)
        return static_code, dynamic_code


class RNNCovariateEncoder(RNNEncoder):
    """
    Deprecated class only for compatibility; use RNNEncoder instead.
    """

    @validated()
    def __init__(
        self,
        use_static_feat: bool = True,
        use_dynamic_feat: bool = True,
        **kwargs,
    ) -> None:
        super().__init__(
            use_static_feat=use_static_feat,
            use_dynamic_feat=use_dynamic_feat,
            **kwargs,
        )
